/*
 * TNET		A server program for MINIX which implements the TCP/IP
 *		suite of networking protocols.  It is based on the
 *		TCP/IP code written by Phil Karn et al, as found in
 *		his NET package for Packet Radio communications.
 *
 *		Create a resolver query record, send it to a nameto a
 *		name server and wait for its answer to arrive.
 *
 * Version:	@(#)inet/res_query.c	1.00		07/11/92
 *
 * Authors:	Original taken from BSD 4.3/TAHOE.
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 */
#include <inet/socket.h>
#include <inet/in.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <netdb.h>
#include <stdio.h>


#ifdef NETWORKED


#if PACKETSZ > 1024
#   define MAXPACKET	PACKETSZ
#else
#   define MAXPACKET	1024
#endif


_PROTOTYPE( char *hostalias, (char *name)				);
_PROTOTYPE( int strcamecmp, (char *, char *)				);


static int strcasecmp(s1, s2)
register char *s1, *s2;
{
  char buf1[128];
  char buf2[128];
  register char *sp;

  sp = buf1;
  while (*s1) {
	if (*s1 >= 'a' && *s1 <= 'z') *sp++ = *s1 - 'a' + 'A';
	  else *sp++ = *s1;
	*s1++;
  }
  *sp = '\0';
  sp = buf2;
  while (*s2) {
	if (*s2 >= 'a' && *s2 <= 'z') *sp++ = *s2 - 'a' + 'A';
	  else *sp++ = *s2;
	*s2++;
  }
  *sp = '\0';
  return(strcmp(buf1, buf2));
}


/*
 * Formulate a normal query, send, and await answer.
 * Returned answer is placed in supplied buffer "answer".
 * Perform preliminary check of answer, returning success only
 * if no error is indicated and the answer count is nonzero.
 * Return the size of the response on success, -1 on error.
 * Error number is left in h_errno.
 * Caller must parse answer and determine whether it answers the question.
 */
int res_query(name, class, type, answer, anslen)
char *name;		/* domain name */
int class, type;	/* class and type of query */
u_char *answer;		/* buffer to put answer */
int anslen;		/* size of answer buffer */
{
  char buf[MAXPACKET];
  HEADER *hp;
  int n;

  if ((_res.options & RES_INIT) == 0 && res_init() == -1) return (-1);
#ifdef DEBUG
  if (_res.options & RES_DEBUG)
	printf("res_query(%s, %d, %d)\n", name, class, type);
#endif
  n = res_mkquery(QUERY, name, class, type, (char *)NULL, 0, NULL,
							buf, sizeof(buf));

  if (n <= 0) {
#ifdef DEBUG
	if (_res.options & RES_DEBUG)
		printf("res_query: mkquery failed\n");
#endif
	h_errno = NO_RECOVERY;
	return (n);
  }
  n = res_send(buf, n, answer, anslen);
  if (n < 0) {
#ifdef DEBUG
	if (_res.options & RES_DEBUG)
		printf("res_query: send error\n");
#endif
	h_errno = TRY_AGAIN;
	return(n);
  }

  hp = (HEADER *) answer;
  if ((ntohs(hp->fourth) & FOURTH_RCODE) != NOERROR ||
	ntohs(hp->ancount) == 0) {
#ifdef DEBUG
	if (_res.options & RES_DEBUG)
		printf("rcode = %d, ancount=%d\n",
			(ntohs(hp->fourth) & FOURTH_RCODE),
					ntohs(hp->ancount));
#endif
	switch ((ntohs(hp->fourth) & FOURTH_RCODE)) {
		case NXDOMAIN:
			h_errno = HOST_NOT_FOUND;
			break;
		case SERVFAIL:
			h_errno = TRY_AGAIN;
			break;
		case NOERROR:
			h_errno = NO_DATA;
			break;
		case FORMERR:
		case NOTIMP:
		case REFUSED:
		default:
			h_errno = NO_RECOVERY;
			break;
	}
	return (-1);
  }
  return(n);
}

/*
 * Formulate a normal query, send, and retrieve answer in supplied buffer.
 * Return the size of the response on success, -1 on error.
 * If enabled, implement search rules until answer or unrecoverable failure
 * is detected.  Error number is left in h_errno.
 * Only useful for queries in the same name hierarchy as the local host
 * (not, for example, for host address-to-name lookups in domain in-addr.arpa).
 */
int res_search(name, class, type, answer, anslen)
char *name;		/* domain name */
int class, type;	/* class and type of query */
u_char *answer;		/* buffer to put answer */
int anslen;		/* size of answer */
{
  register char *cp, **domain;
  int n, ret, got_nodata = 0;
  char *hostalias();

  if ((_res.options & RES_INIT) == 0 && res_init() == -1) return (-1);

  errno = 0;
  h_errno = HOST_NOT_FOUND;		/* default, if we never query */
  for (cp = name, n = 0; *cp; cp++)
	if (*cp == '.') n++;
  if (n == 0 && (cp = hostalias(name)))
	return (res_query(cp, class, type, answer, anslen));

  /*
   * We do at least one level of search if
   *	- there is no dot and RES_DEFNAME is set, or
   *	- there is at least one dot, there is no trailing dot,
   *	  and RES_DNSRCH is set.
   */
  if ((n == 0 && _res.options & RES_DEFNAMES) ||
	(n != 0 && *--cp != '.' && _res.options & RES_DNSRCH))
	     for (domain = _res.dnsrch; *domain; domain++) {
		ret = res_qdomain(name, *domain, class, type,
		    answer, anslen);
		if (ret > 0) return (ret);
		/*
		 * If no server present, give up.
		 * If name isn't found in this domain,
		 * keep trying higher domains in the search list
		 * (if that's enabled).
		 * On a NO_DATA error, keep trying, otherwise
		 * a wildcard entry of another type could keep us
		 * from finding this entry higher in the domain.
		 * If we get some other error (negative answer or
		 * server failure), then stop searching up,
		 * but try the input name below in case it's fully-qualified.
		 */
		if (errno == ECONNREFUSED) {
			h_errno = TRY_AGAIN;
			return (-1);
		}
		if (h_errno == NO_DATA)
			got_nodata++;
		if ((h_errno != HOST_NOT_FOUND && h_errno != NO_DATA) ||
		    (_res.options & RES_DNSRCH) == 0)
			break;
	}
	/*
	 * If the search/default failed, try the name as fully-qualified,
	 * but only if it contained at least one dot (even trailing).
	 * This is purely a heuristic; we assume that any reasonable query
	 * about a top-level domain (for servers, SOA, etc) will not use
	 * res_search.
	 */
	if (n && (ret = res_qdomain(name, (char *)NULL, class, type,
	    answer, anslen)) > 0)
		return (ret);
	if (got_nodata)
		h_errno = NO_DATA;
	return (-1);
}


/*
 * Perform a call on res_query on the concatenation of name and domain,
 * removing a trailing dot from name if domain is NULL.
 */
int res_qdomain(name, domain, class, type, answer, anslen)
char *name, *domain;
int class, type;	/* class and type of query */
u_char *answer;		/* buffer to put answer */
int anslen;		/* size of answer */
{
  char nbuf[2*MAXDNAME+2];
  char *longname = nbuf;
  int n;

#ifdef DEBUG
  if (_res.options & RES_DEBUG)
	printf("res_qdomain(%s, %s, %d, %d)\n",
				name, domain, class, type);
#endif
  if (domain == NULL) {
	/*
	 * Check for trailing '.';
	 * copy without '.' if present.
	 */
	n = strlen(name) - 1;
	if (name[n] == '.' && n < sizeof(nbuf) - 1) {
		bcopy(name, nbuf, n);
		nbuf[n] = '\0';
	} else longname = name;
  } else (void)sprintf(nbuf, "%.*s.%.*s", MAXDNAME, name, MAXDNAME, domain);

  return(res_query(longname, class, type, answer, anslen));
}


char *hostalias(name)
register char *name;
{
  register char *C1, *C2;
  FILE *fp;
  char *file, *getenv(), *strcpy(), *strncpy();
  char buf[BUFSIZ];
  static char abuf[MAXDNAME];

  file = getenv("HOSTALIASES");
  if (file == NULL || (fp = fopen(file, "r")) == NULL) return(NULL);
  buf[sizeof(buf) - 1] = '\0';
  while (fgets(buf, sizeof(buf), fp)) {
	for (C1 = buf; *C1 && !isspace(*C1); ++C1);
	if (!*C1) break;
	*C1 = '\0';
	if (!strcasecmp(buf, name)) {
		while (isspace(*++C1));
		if (!*C1) break;
		for (C2 = C1 + 1; *C2 && !isspace(*C2); ++C2);
		abuf[sizeof(abuf) - 1] = *C2 = '\0';
		(void)strncpy(abuf, C1, sizeof(abuf) - 1);
		fclose(fp);
		return(abuf);
	}
  }
  fclose(fp);
  return(NULL);
}
#endif	/* NETWORKED */
